@@ -23,22 +23,37 @@ import filecmp
 import unicodedata
 import codecs
 
-__version__ = "1.9.1"
+__version__ = "1.9.3"
 
 color_codes = {
+    "black":   '\033[0;30m',
     "red":     '\033[0;31m',
     "green":   '\033[0;32m',
     "yellow":  '\033[0;33m',
     "blue":    '\033[0;34m',
     "magenta": '\033[0;35m',
     "cyan":    '\033[0;36m',
+    "white":   '\033[0;37m',
     "none":    '\033[m',
+    "black_bold":   '\033[1;30m',
     "red_bold":     '\033[1;31m',
     "green_bold":   '\033[1;32m',
     "yellow_bold":  '\033[1;33m',
     "blue_bold":    '\033[1;34m',
     "magenta_bold": '\033[1;35m',
     "cyan_bold":    '\033[1;36m',
+    "white_bold":   '\033[1;37m',
+}
+
+
+color_mapping = {
+    "add": "green_bold",
+    "subtract": "red_bold",
+    "change": "yellow_bold",
+    "separator": "blue",
+    "description": "blue",
+    "meta": "magenta",
+    "line-numbers": "white",
 }
 
 
@@ -72,7 +87,6 @@ class ConsoleDiff(object):
                  line_numbers=False,
                  show_all_spaces=False,
                  highlight=False,
-                 no_bold=False,
                  strip_trailing_cr=False):
         """ConsoleDiff instance initializer
 
@@ -90,7 +104,6 @@ class ConsoleDiff(object):
         self.cols = cols
         self.show_all_spaces = show_all_spaces
         self.highlight = highlight
-        self.no_bold = no_bold
         self.strip_trailing_cr = strip_trailing_cr
 
         if wrapcolumn is None:
@@ -259,7 +272,8 @@ class ConsoleDiff(object):
             # handle blank lines where linenum is '>' or ''
             lid = ''
             return text
-        return '%s %s' % (self._rpad(lid, 8), text)
+        return '%s %s' % (
+            self._rpad(simple_colorize(str(lid), "line-numbers"), 8), text)
 
     def _real_len(self, s):
         l = 0
@@ -340,32 +354,27 @@ class ConsoleDiff(object):
 
     def _generate_table(self, fromdesc, todesc, diffs):
         if fromdesc or todesc:
-            yield (simple_colorize(fromdesc, "blue"),
-                   simple_colorize(todesc, "blue"))
+            yield (simple_colorize(fromdesc, "description"),
+                   simple_colorize(todesc, "description"))
 
         for i, line in enumerate(diffs):
             if line is None:
                 # mdiff yields None on separator lines; skip the bogus ones
                 # generated for the first line
                 if i > 0:
-                    yield (simple_colorize('---', "blue"),
-                           simple_colorize('---', "blue"))
+                    yield (simple_colorize('---', "separator"),
+                           simple_colorize('---', "separator"))
             else:
                 yield line
 
     def colorize(self, s):
         def background(color):
-            return replace_all({"\033[1;": "\033[7;",
+            return replace_all({"\033[1;": "\033[7;1;",
                                 "\033[0;": "\033[7;"}, color)
 
-        if self.no_bold:
-            C_ADD = color_codes["green"]
-            C_SUB = color_codes["red"]
-            C_CHG = color_codes["yellow"]
-        else:
-            C_ADD = color_codes["green_bold"]
-            C_SUB = color_codes["red_bold"]
-            C_CHG = color_codes["yellow_bold"]
+        C_ADD = color_codes[color_mapping["add"]]
+        C_SUB = color_codes[color_mapping["subtract"]]
+        C_CHG = color_codes[color_mapping["change"]]
 
         if self.highlight:
             C_ADD, C_SUB, C_CHG = (background(C_ADD),
@@ -388,7 +397,7 @@ class ConsoleDiff(object):
         if not self.show_all_spaces:
             # If there's a change consisting entirely of whitespace,
             # don't color it.
-            return re.sub("\033\\[[01];3([123])m(\\s+)(\033\\[)",
+            return re.sub("\033\\[[01];3([01234567])m(\\s+)(\033\\[)",
                           "\033[7;3\\1m\\2\\3", s)
 
         def will_see_coloredspace(i, s):
@@ -426,8 +435,12 @@ class ConsoleDiff(object):
         return joined
 
 
-def simple_colorize(s, chosen_color):
-    return "%s%s%s" % (color_codes[chosen_color], s, color_codes["none"])
+def raw_colorize(s, color):
+    return "%s%s%s" % (color_codes[color], s, color_codes["none"])
+
+
+def simple_colorize(s, category):
+    return raw_colorize(s, color_mapping[category])
 
 
 def replace_all(replacements, string):
@@ -461,6 +474,12 @@ def create_option_parser():
                       "Unix only")
     parser.add_option("--encoding", default="utf-8",
                       help="specify the file encoding; defaults to utf8")
+    parser.add_option("-E", "--exclude-lines",
+                      action="store",
+                      type="string",
+                      dest='matcher',
+                      help="Do not diff lines that match this regex. Not "
+                      "compatible with the 'line-numbers' option")
     parser.add_option("--head", default=0,
                       help="consider only the first N lines of each file")
     parser.add_option("-H", "--highlight", default=False,
@@ -476,7 +495,8 @@ def create_option_parser():
                       "Use twice, one for each file")
     parser.add_option("-N", "--line-numbers", default=False,
                       action="store_true",
-                      help="generate output with line numbers")
+                      help="generate output with line numbers. Not compatible "
+                      "with the 'exclude-lines' option.")
     parser.add_option("--no-bold", default=False,
                       action="store_true",
                       help="use non-bold colors; recommended for solarized")
@@ -512,6 +532,13 @@ def create_option_parser():
                       action="store_true",
                       help="strip any trailing carriage return at the end of "
                       "an input line")
+    parser.add_option("--color-map", default=None,
+                      help="choose which colors are used for which items. "
+                      "Default is --color-map='" +
+                      ",".join("%s:%s" % x
+                               for x in sorted(color_mapping.items())) + "'"
+                      ".  You don't have to override all of them: "
+                      "'--color-map=separator:white,description:cyan")
     return parser
 
 
@@ -528,7 +555,7 @@ def set_cols_option(options):
                 return None
             return cr
         cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
-        if cr:
+        if cr and cr[1] > 0:
             options.cols = cr[1]
         else:
             options.cols = 80
@@ -559,7 +586,7 @@ def codec_print(s, options):
 
 def diff(options, a, b):
     def print_meta(s):
-        codec_print(simple_colorize(s, "magenta"), options)
+        codec_print(simple_colorize(s, "meta"), options)
 
     # Don't use os.path.isfile; it returns False for file-like entities like
     # bash's process substitution (/dev/fd/N).
@@ -635,10 +662,40 @@ def diff_files(options, a, b):
         lines_a = lines_a[:head]
         lines_b = lines_b[:head]
 
+    if options.matcher:
+        lines_a = [line_a for line_a in lines_a if
+                   not re.search(options.matcher, line_a)]
+        lines_b = [line_b for line_b in lines_b if
+                   not re.search(options.matcher, line_b)]
+
+    if options.no_bold:
+        for key in color_mapping:
+            color_mapping[key] = color_mapping[key].replace("_bold", "")
+
+    if options.color_map:
+        command_for_errors = '--color-map="%s"' % (options.color_map)
+        for mapping in options.color_map.split(","):
+            category, color = mapping.split(":", 1)
+
+            if category not in color_mapping:
+                print(
+                    "Invalid category '%s' in '%s'.  Valid categories are: %s."
+                    % (category, command_for_errors,
+                       ", ".join(sorted(color_mapping.keys()))))
+                sys.exit()
+
+            if color not in color_codes:
+                print("Invalid color '%s' in '%s'.  Valid colors are: %s."
+                      % (color, command_for_errors,
+                         ", ".join([raw_colorize(x, x)
+                                    for x in sorted(color_codes.keys())])))
+                sys.exit()
+
+            color_mapping[category] = color
+
     cd = ConsoleDiff(cols=int(options.cols),
                      show_all_spaces=options.show_all_spaces,
                      highlight=options.highlight,
-                     no_bold=options.no_bold,
                      line_numbers=options.line_numbers,
                      tabsize=int(options.tabsize),
                      strip_trailing_cr=options.strip_trailing_cr)
